/*
 * Copyright (C) 2004 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ruby.h>

#include <Evas.h>

#include "rb_evas_main.h"
#include "rb_evas.h"
#include "rb_evas_object.h"

#define SMART_CB_BODY(name) \
	VALUE self = TO_EVAS_OBJECT (o); \
	static ID id; \
\
	if (!id) \
		id = rb_intern ("smart_"#name); \
\
	if (!rb_respond_to (self, id)) \
		return;

#define SMART_CB(name) \
	static void smart_##name (Evas_Object *o) \
	{ \
		SMART_CB_BODY (name); \
		rb_funcall (self, id, 0); \
	}

#define SMART_CB_OBJ(name) \
	static void smart_##name (Evas_Object *o, Evas_Object *other) \
	{ \
		SMART_CB_BODY (name); \
		rb_funcall (self, id, 1, TO_EVAS_OBJECT (other)); \
	}

#define SMART_CB_COORD(name) \
	static void smart_##name (Evas_Object *o, Evas_Coord a, Evas_Coord b) \
	{ \
		SMART_CB_BODY (name); \
		rb_funcall (self, id, 2, INT2FIX ((int) a), INT2FIX ((int) b)); \
	}

typedef struct {
	Evas_Smart_Class smart_class;
	Evas_Smart *smart;

	VALUE name;
} SmartData;

static ID id_smart_data;

SMART_CB (delete);
SMART_CB (show);
SMART_CB (hide);
SMART_CB (clip_unset);
SMART_CB_OBJ (clip_set);
SMART_CB_COORD (move);
SMART_CB_COORD (resize);

static void smart_color_set (Evas_Object *o, int r, int g, int b, int a)
{
	SMART_CB_BODY (color_set);

	rb_funcall (self, id, 4, INT2FIX (r), INT2FIX (g), INT2FIX (b),
	            INT2FIX (a));
}

static VALUE c_inherited (VALUE klass, VALUE child)
{
	rb_const_set (child, id_smart_data, Qnil);

	return Qnil;
}

static void
c_data_mark (SmartData *sd)
{
	rb_gc_mark (sd->name);
}

static VALUE c_init (VALUE self, VALUE evas)
{
	VALUE klass, data;
	SmartData *sd;

	CHECK_CLASS (evas, cEvas);
	GET_OBJ (evas, RbEvas, e);
	GET_OBJ (self, RbEvasObject, s2);

	klass = rb_obj_class (self);

	/* check whether the smart object has been created already */
	data = rb_const_get (klass, id_smart_data);
	if (!NIL_P (data))
		Data_Get_Struct (data, SmartData, sd);
	else {
		data = Data_Make_Struct (rb_cObject, SmartData,
		                         c_data_mark, NULL, sd);

		sd->name = rb_class_path (klass);

		sd->smart_class.name = StringValuePtr (sd->name);
		sd->smart_class.version = EVAS_SMART_CLASS_VERSION;

		sd->smart_class.add = NULL;
		sd->smart_class.del = smart_delete;
		sd->smart_class.move = smart_move;
		sd->smart_class.resize = smart_resize;
		sd->smart_class.show = smart_show;
		sd->smart_class.hide = smart_hide;
		sd->smart_class.color_set = smart_color_set;
		sd->smart_class.clip_set = smart_clip_set;
		sd->smart_class.clip_unset = smart_clip_unset;

		sd->smart = evas_smart_class_new (&sd->smart_class);

		rb_mod_remove_const (klass, ID2SYM (id_smart_data));
		rb_const_set (klass, id_smart_data, data);
	}

	s2->real = evas_object_smart_add (e->real, sd->smart);

	rb_call_super (1, &evas);

	return self;
}

static VALUE c_add_member (VALUE self, VALUE member)
{
	GET_OBJ (self, RbEvasObject, e);
	GET_OBJ (member, RbEvasObject, e2);

	/* weird order of arguments */
	evas_object_smart_member_add (e2->real, e->real);

	return Qnil;
}

void Init_Smart (void)
{
	VALUE c = rb_define_class_under (mEvas, "Smart", cEvasObject);

	rb_define_singleton_method (c, "inherited", c_inherited, 1);
	rb_define_method (c, "initialize", c_init, 1);
	rb_define_method (c, "add_member", c_add_member, 1);

	id_smart_data = rb_intern ("SMART_DATA");
}
